/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Ahmad Boorghany
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <cmath>
#include <cassert>

#include <Logger.h>
#include <Degree.h>
#include <Defines.h>
#include <WorldModel.h>
#include <InterceptCalculate.h>


using namespace std;
using namespace Degree;

// class InterceptPlayer

InterceptPlayer::InterceptPlayer()
{
	player = NULL;
	withTurnVel = NULL;
	withoutTurnVel = NULL;
	withTurnDashDist = NULL;
	withoutTurnDashDist = NULL;
}

void InterceptPlayer::simulate()
{
	withTurnVel++;
	withTurnDashDist++;

	withoutTurnVel++;
	withoutTurnDashDist++;
}

void InterceptPlayer::getReadyToRun()
{
	// Reseting dynamic variables
	setCatchBall(false);

	withTurnVel = withTurnVelArchive;
	withTurnDashDist = withTurnDashDistArchive;

	withoutTurnVel = withoutTurnVelArchive;
	withoutTurnDashDist = withoutTurnDashDistArchive;
}

void InterceptPlayer::init(const Player &playerArg, const Ball &ball,
		bool kickingFlagArg, unsigned kickCycles,
		unsigned seeBallDelayArg, float slowDashPower,
		float fastDashPower, float directDashPower,
		float posDevDashDist, bool blockedPlayer, bool checkForPenaltyBox)
{
	unsigned cyclesCounter;

	blockedFlag = blockedPlayer;

	player = &playerArg;
	simPlayer = *player;

//	Body tempBody; // A Theory says these lines are bug.
//	simPlayer.simulateByDynamics(tempBody);

    if (player->isGoalie())
	{
		if (!checkForPenaltyBox ||
			(player->getTeamId() == TID_TEAMMATE &&
			 player->getPos().getX() < -36.4 &&
			 fabs(player->getPos().getY()) < 19.6) ||
			(player->getTeamId() == TID_OPPONENT &&
			 player->getPos().getX() > 36.4 &&
			 fabs(player->getPos().getY()) < 19.6))
			kickableRadius = player->getCatchableAreaL();
		else
			kickableRadius = player->getSize() +
					player->getKickableMargin() +
					ball.getSize();
	}
	else
		kickableRadius = player->getSize() +
				player->getKickableMargin() +
				ball.getSize();

	slowAccel = slowDashPower * player->getDashPowerRate() *
				player->getEffortMax();
	fastAccel = fastDashPower * player->getDashPowerRate() *
				player->getEffortMax();
	directAccel = directDashPower * player->getDashPowerRate() *
				player->getEffortMax();

	seeBallDelay = seeBallDelayArg;
	setKicking(kickingFlagArg);

	// Setting start values
	withTurnVelArchive[0] = 0;
	withTurnDashDistArchive[0] = posDevDashDist;

	withoutTurnVelArchive[0] = player->getVel().getMagnitude();
	withoutTurnDashDistArchive[0] = posDevDashDist;

	if (isLogging())
		LOG << "NewP: " << player->getTeamId() << " " << player->getUniNum()
			<< " Pos" << simPlayer.getPos() << " ----------" << endl
			<< "WithTurnDashVel[0]: " << withTurnDashDistArchive[0] << " "
			<< withTurnVelArchive[0] << endl
			<< "NoTurnDashVel[0]: " << withoutTurnDashDistArchive[0] << " "
			<< withoutTurnVelArchive[0] << endl;

	// Setting other values
	for (cyclesCounter = 1; cyclesCounter < INCA_MAX_INTERCEPT_TIME; cyclesCounter++)
	{
		// Copying values from last cycle
		withTurnDashDistArchive[cyclesCounter] =
				withTurnDashDistArchive[cyclesCounter - 1];
		withoutTurnDashDistArchive[cyclesCounter] =
				withoutTurnDashDistArchive[cyclesCounter - 1];

		withTurnVelArchive[cyclesCounter] =
				withTurnVelArchive[cyclesCounter - 1];
		withoutTurnVelArchive[cyclesCounter] =
				withoutTurnVelArchive[cyclesCounter - 1];

		// Accelerating
		if (!isKicking() &&
			cyclesCounter <= seeBallDelay + kickCycles) // The ball is kicking or I still don't know it has kicked.
		{
			withTurnVelArchive[cyclesCounter] += slowAccel;
			withoutTurnVelArchive[cyclesCounter] += directAccel;

			if (withTurnVelArchive[cyclesCounter] > player->getSpeedMax())
				withTurnVelArchive[cyclesCounter] = player->getSpeedMax();
			if (withoutTurnVelArchive[cyclesCounter] > player->getSpeedMax())
				withoutTurnVelArchive[cyclesCounter] = player->getSpeedMax();
		}
		else
		{
			if (cyclesCounter > seeBallDelay + kickCycles)
			{
				withoutTurnVelArchive[cyclesCounter] += fastAccel;
				if (withoutTurnVelArchive[cyclesCounter] > player->getSpeedMax())
					withoutTurnVelArchive[cyclesCounter] = player->getSpeedMax();
			}

			if (cyclesCounter > seeBallDelay + kickCycles + 1)
			{
				withTurnVelArchive[cyclesCounter] += fastAccel;
				if (withTurnVelArchive[cyclesCounter] > player->getSpeedMax())
					withTurnVelArchive[cyclesCounter] = player->getSpeedMax();
			}
		}

		// Moving
		withTurnDashDistArchive[cyclesCounter] += 
				withTurnVelArchive[cyclesCounter];
		withoutTurnDashDistArchive[cyclesCounter] += 
				withoutTurnVelArchive[cyclesCounter];

		// Simulating
		withTurnVelArchive[cyclesCounter] *= player->getDecay();
		withoutTurnVelArchive[cyclesCounter] *= player->getDecay();
	}
}

void InterceptPlayer::finalizeIntercept(const Ball &ball, unsigned interceptTimeArg)
{
	setCatchBall(true);

	interceptTime = interceptTimeArg;
	interceptPoint = ball.getPos();
}

bool InterceptPlayer::isBallKickable(const Ball &ball)
{
	if (player->getDistance(ball) <= kickableRadius)
		return true;
	return false;
}

bool InterceptPlayer::checkCatchBall(const Ball &ball)
{
	bool flag = false;
	float ballDist;
	float deltaAngle;
	float validDeltaAngle;
	Vector vector;

	if (isLogging())
		LOG << "Player: " << player->getTeamId() << " " << player->getUniNum()
			<< " Pos" << simPlayer.getPos() << " ----------" << endl
			<< "WithTurnDashVel: " << *withTurnDashDist << " "
			<< *withTurnVel << endl
			<< "NoTurnDashVel: " << *withoutTurnDashDist << " "
			<< *withoutTurnVel << endl;

	// Without turn catch ball check
	ballDist = simPlayer.getDistance(ball) - *withoutTurnDashDist;

	if (isLogging())
		LOG << "NoTurnBallDist: " << ballDist << endl;

	if (ballDist < kickableRadius)
	{
		if (player->getBodyDir() != NOVALUE)
		{
			vector.setByPoints(simPlayer.getPos(), ball.getPos());
			validDeltaAngle = arcTan(kickableRadius / vector.getMagnitude());

			deltaAngle = getDeltaAngle(vector.getDirection(),
					simPlayer.getBodyDir());

			if (fabs(deltaAngle) <= validDeltaAngle ||
				fabs(deltaAngle) <= INCA_MIN_DELTA_ANGLE)
				flag = true;
			else // Checking back dash.
			{
				deltaAngle = getDeltaAngle(vector.getDirection(),
						simPlayer.getBodyDir() + 180);

				if (fabs(deltaAngle) <= validDeltaAngle ||
					fabs(deltaAngle) <= INCA_MIN_DELTA_ANGLE)
					flag = true;
			}
		}

		// With turn catch ball check
		if (flag == false)
		{
			ballDist = simPlayer.getDistance(ball) - *withTurnDashDist;

			if (isLogging())
				LOG << "WithTurnBallDist: " << ballDist << endl;

			if (ballDist < kickableRadius)
				flag = true;
		}
	}

	if (flag)
	{

		if (isLogging())
			LOG << "*** I catched the ball ***" << endl;

		return true;
	}

	return false;
}

void InterceptPlayer::checkForExtraCatch(const Ball &ball)
{
	float ballDist = simPlayer.getDistance(ball) - *withTurnDashDist;

	if (isLogging())
		LOG << "Player: " << player->getTeamId() << " " << player->getUniNum()
			<< " Pos" << simPlayer.getPos() << " ----EC----" << endl
			<< "WithTurnDashVel: " << *withTurnDashDist << " "
			<< *withTurnVel << endl
			<< "WithTurnBallDist: " << ballDist << endl;

	if (ballDist < kickableRadius)
	{
		extraCatchedFlag = true;

		if (isLogging())
			LOG << "*** I extra catched the ball ***" << endl;
	}
}

// Getting functions
bool InterceptPlayer::isLogging() const
{
	return loggingFlag;
}

bool InterceptPlayer::isKicking() const
{
	return kickingFlag;
}

bool InterceptPlayer::isBlocked() const
{
	return blockedFlag;
}

bool InterceptPlayer::isCatchBall() const
{
	return catchBallFlag;
}

bool InterceptPlayer::isExtraCatched() const
{
	return extraCatchedFlag;
}

float InterceptPlayer::getDashDist() const
{
	return *withTurnDashDist;
}

unsigned InterceptPlayer::getInterTime() const
{
	return interceptTime;
}

Point InterceptPlayer::getInterPoint() const
{
	return interceptPoint;
}

const Player &InterceptPlayer::getPlayer() const
{
	return *player;
}

float InterceptPlayer::getKickableRadius() const
{
	return kickableRadius;
}

// Setting functions
void InterceptPlayer::setLogging(bool loggingFlagArg)
{
	loggingFlag = loggingFlagArg;
}

void InterceptPlayer::setKicking(bool kickingFlagArg)
{
	kickingFlag = kickingFlagArg;
}

void InterceptPlayer::setCatchBall(bool catchBallFlagArg)
{
	catchBallFlag = catchBallFlagArg;
}

void InterceptPlayer::setExtraCatched(bool extraCatchedFlagArg)
{
	extraCatchedFlag = extraCatchedFlagArg;
}

void InterceptPlayer::setKickableRadius(float kickableRadiusArg)
{
	kickableRadius = kickableRadiusArg;
}


// class InterceptCalculate

void InterceptCalculate::run()
{
	unsigned finishTimer = NOVALUE;
	unsigned interceptsCycle = 0;
	unsigned kickCyclesCounter = kickCycles;
	unsigned i;

	if (isLogging())
		LOG << "Running Intercepts." << endl;

	getReadyToRun();

	// checking the moment
	if (kickCyclesCounter == 0)
		for (i = 0; i < playersNum; i++)
			if (interceptPlayers[i].isBallKickable(virtualBall))
			{
				interceptPlayers[i].finalizeIntercept(virtualBall,
						interceptsCycle);

				if (!interceptPlayers[i].isBlocked())
				{
					checkedPlayers[checkedPlayersNum] = i;
					checkedPlayersNum++;
					if (finishTimer == NOVALUE)
						finishTimer = 0;

					if (checkedPlayersNum == MAX_CHECK_PLAYERS ||
						interceptPlayers[i].getPlayer().isGoalie())
					{
						finalizeIntercepts();
						return;
					}
				}
				else
					setBlockedBall();
			}

	// Running GoToBalls
	while (1)
	{
		interceptsCycle++;
		if (finishTimer != NOVALUE)
			finishTimer++;

		if (isLogging())
			LOG << "Time: " << interceptsCycle << " - "
				<< "FinishTime: " << finishTimer << " ---------------------"
				<< endl	<< "Simulating objects..." << endl;

		// simulating the ball and players
		for (i = 0; i < playersNum; i++)
			if (!interceptPlayers[i].isCatchBall())
				interceptPlayers[i].simulate();

		if (kickCyclesCounter > 0)
			kickCyclesCounter--;
		if (kickCyclesCounter == 0)
			simulateVirtualBall();

		if (isLogging())
			LOG << "Ball:"
				<< " Pos" << virtualBall.getPos()
				<< " Vel" << virtualBall.getVel() << endl
				<< "Checking for catches..." << endl;

		// Checking players for catching the ball
		for (i = 0; i < playersNum; i++)
 			if (!interceptPlayers[i].isCatchBall())
			{
				if (interceptPlayers[i].checkCatchBall(virtualBall))
				{
					interceptPlayers[i].finalizeIntercept(virtualBall,
						interceptsCycle);

					if (!interceptPlayers[i].isBlocked())
					{
						checkedPlayers[checkedPlayersNum] = i;
						checkedPlayersNum++;
						if (finishTimer == NOVALUE)
							finishTimer = 0;

						if (checkedPlayersNum == MAX_CHECK_PLAYERS ||
							interceptPlayers[i].getPlayer().isGoalie())
						{
							finalizeIntercepts();
							return;
						}
					}
					else
						setBlockedBall();
				}

			}
			else if (!interceptPlayers[i].isExtraCatched())
				interceptPlayers[i].checkForExtraCatch(virtualBall);

		if (finishTimer != NOVALUE &&
			finishTimer >= shareCycles)
		{
			if (isLogging())
				LOG << "*** Finish : finishTimer overflow ***" << endl;

			finalizeIntercepts();
			return;
		}

		if (interceptsCycle > INCA_MAX_INTERCEPT_TIME - 5)
		{
			if (isLogging())
				LOG << "*** Finish : interceptCycles overflow ***" << endl;

			finalizeIntercepts();
			return;
		}

		if (playersNum == 0 &&
			virtualBall.getVel().getMagnitude() < 0.05) // the ball is stoped in an empty field.
		{
			if (isLogging())
				LOG << "*** Finish : Ownerless ball in the field ***" << endl;

			finalizeIntercepts();
			return;
		}

		if (WorldModel::getOutDistance(virtualBall.getPos()) > INCA_OUT_DIST)
		{
			if (isLogging())
				LOG << "*** Finish : The ball goes out ***" << endl;

			playMode = PM_KICK_OFF_OPP;
			finalizeIntercepts();
			return;
		}
	}
}

void InterceptCalculate::prepareSum()
{
	if (checkedPlayersNum == 0)
		return;

	fastestCheckedPlayerNum = 0;
	fastestPlayerNum = checkedPlayers[0];

	for (unsigned i = 0; i < checkedPlayersNum; i++)
		if (getCheckedInterPlayer(i).getPlayer().getTeamId() == TID_TEAMMATE)
		{
			fastestTmmCheckedPlayerNum = i;
			fastestTmmPlayerNum = checkedPlayers[i];
			break;
		}

	for (unsigned i = 0; i < checkedPlayersNum; i++)
		if (getCheckedInterPlayer(i).getPlayer().getTeamId() == TID_OPPONENT)
		{
			fastestOppCheckedPlayerNum = i;
			fastestOppPlayerNum = checkedPlayers[i];
			break;
		}
}

void InterceptCalculate::initTeamsProps()
{
	if (kickCycles > 0)
	{
		kickerSeeBallDelay = 0;

		tmmPosDevRate = -0.75;
		tmmSeeBallDelay = 2;
		tmmGoalieSeeBallDelay = 0;
		tmmSlowDashPower = 0;
		tmmFastDashPower = 100;
		tmmDirectDashPower = 0;
		tmmGoalieDirectDashPower = 0;

		oppPosDevRate = 0.5;
		oppSeeBallDelay = 0;
		oppGoalieSeeBallDelay = 0;
		oppSlowDashPower = 45;
		oppFastDashPower = 100;
		oppDirectDashPower = 60;
		oppGoalieDirectDashPower = 100;
	}
	else
	{
		kickerSeeBallDelay = 0; // only for reseting

		tmmPosDevRate = -0.3;
		tmmSeeBallDelay = 0;
		tmmGoalieSeeBallDelay = 0;
		tmmSlowDashPower = 30;
		tmmFastDashPower = 100;
		tmmDirectDashPower = 30;
		tmmGoalieDirectDashPower = 30;

		oppPosDevRate = 0.3;
		oppSeeBallDelay = 0;
		oppGoalieSeeBallDelay = 0;
		oppSlowDashPower = 50;
		oppFastDashPower = 100;
		oppDirectDashPower = 90;
		oppGoalieDirectDashPower = 100;
	}
}

void InterceptCalculate::startSession(unsigned kickCyclesArg,
		const Player *kickPlayerArg, const Player *blockedPlayerArg)
{
	setLogging(false);
	playersNum = 0;
	kickPlayer = kickPlayerArg;
	blockedPlayer = blockedPlayerArg;
	kickCycles = kickCyclesArg;
	shareCycles = DEFAULT_SHARE_CYCLES;

	initTeamsProps();
}

void InterceptCalculate::addPlayer(const Player &player, bool kickPlayerFlag,
		bool blockedPlayerFlag, bool checkForPenaltyBox)
{
	float posDevDashDist;

	interceptPlayers[playersNum].setLogging(isLogging());

	if (kickPlayerFlag)
		interceptPlayers[playersNum].init(player, virtualBall,
				kickPlayerFlag, kickCycles, kickerSeeBallDelay,
				0, player.getMaxPower(), 0, 0, false, checkForPenaltyBox);

	else if (player.getTeamId() == TID_TEAMMATE)
	{
		posDevDashDist = tmmPosDevRate * player.getPosDeviation();

		if (player.isGoalie())
			interceptPlayers[playersNum].init(player, virtualBall,
					kickPlayerFlag, kickCycles, tmmGoalieSeeBallDelay,
					tmmSlowDashPower, tmmFastDashPower,
					tmmGoalieDirectDashPower, posDevDashDist,
					blockedPlayerFlag, checkForPenaltyBox);
		else
			interceptPlayers[playersNum].init(player, virtualBall,
					kickPlayerFlag, kickCycles, tmmSeeBallDelay,
					tmmSlowDashPower, tmmFastDashPower,
					tmmDirectDashPower, posDevDashDist, blockedPlayerFlag,
					checkForPenaltyBox);
	}

	else if (player.getTeamId() == TID_OPPONENT)
	{
		posDevDashDist = oppPosDevRate * player.getPosDeviation();

		if (player.isGoalie())
			interceptPlayers[playersNum].init(player, virtualBall,
					kickPlayerFlag, kickCycles, oppGoalieSeeBallDelay,
					oppSlowDashPower, oppFastDashPower,
					oppGoalieDirectDashPower, posDevDashDist,
					blockedPlayerFlag, checkForPenaltyBox);
		else
			interceptPlayers[playersNum].init(player, virtualBall,
					kickPlayerFlag, kickCycles, oppSeeBallDelay,
					oppSlowDashPower, oppFastDashPower,
					oppDirectDashPower, posDevDashDist, blockedPlayerFlag,
					checkForPenaltyBox);
	}
	else
		assert(0);

	playersNum++;
}

void InterceptCalculate::updateByWorldModel(const WorldModel *worldModel,
		bool removeBody, bool removeTmm, bool removeOpp)
{
	unsigned i;

	// adding the ball
	virtualBall = worldModel->getBall();

	// adding players
	if (!removeBody)
		addPlayer(worldModel->getBody(),
			(&worldModel->getBody() == kickPlayer),
			(&worldModel->getBody() == blockedPlayer));

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
	{
		if (!removeTmm)
			if (worldModel->getFullPlayer(TID_TEAMMATE, i).isValid() &&
				!worldModel->getFullPlayer(TID_TEAMMATE, i).isBody())
				addPlayer(worldModel->getFullPlayer(TID_TEAMMATE, i),
					(&worldModel->getFullPlayer(TID_TEAMMATE, i) == kickPlayer),
					(&worldModel->getFullPlayer(TID_TEAMMATE, i) == blockedPlayer));

		if (!removeOpp)
			if (worldModel->getFullPlayer(TID_OPPONENT, i).isValid())
				addPlayer(worldModel->getFullPlayer(TID_OPPONENT, i),
					(&worldModel->getFullPlayer(TID_OPPONENT, i) == kickPlayer),
					(&worldModel->getFullPlayer(TID_OPPONENT, i) == blockedPlayer));
	}

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
	{
		if (!removeTmm)
			if (worldModel->getHalfPlayer(TID_TEAMMATE, i).isValid())
				addPlayer(worldModel->getHalfPlayer(TID_TEAMMATE, i),
					(&worldModel->getHalfPlayer(TID_TEAMMATE, i) == kickPlayer),
					(&worldModel->getHalfPlayer(TID_TEAMMATE, i) ==	blockedPlayer));

		if (!removeOpp)
			if (worldModel->getHalfPlayer(TID_OPPONENT, i).isValid())
				addPlayer(worldModel->getHalfPlayer(TID_OPPONENT, i),
					(&worldModel->getHalfPlayer(TID_OPPONENT, i) == kickPlayer),
					(&worldModel->getHalfPlayer(TID_OPPONENT, i) ==	blockedPlayer));
	}
}

void InterceptCalculate::getReadyToRun()
{
	// Resetting after running Intercepts variables
	playMode = PM_PLAY_ON;
	checkedPlayersNum = 0;
	fastestPlayerNum = NOVALUE;
	fastestTmmPlayerNum = NOVALUE;
	fastestOppPlayerNum = NOVALUE;
	fastestCheckedPlayerNum = NOVALUE;
	fastestTmmCheckedPlayerNum = NOVALUE;
	fastestOppCheckedPlayerNum = NOVALUE;

	for (unsigned i = 0;i < playersNum;i++)
		interceptPlayers[i].getReadyToRun();
}

void InterceptCalculate::simulateVirtualBall()
{
	virtualBall.setPos() += virtualBall.getVel();
	virtualBall.setVel() *= virtualBall.getDecay();
}

void InterceptCalculate::finalizeIntercepts()
{
	prepareSum();
}

bool InterceptCalculate::existFastestPlayer() const
{
	if (fastestPlayerNum == NOVALUE)
		return false;
	return true;
}

bool InterceptCalculate::existFastestTmmPlayer() const
{
	if (fastestTmmPlayerNum == NOVALUE)
		return false;
	return true;
}

bool InterceptCalculate::existFastestOppPlayer() const
{
	if (fastestOppPlayerNum == NOVALUE)
		return false;
	return true;
}

bool InterceptCalculate::isFastestPlayer(const Player &player) const
{
	if (getFastestPlayer().getModel() == player.getModel() &&
		getFastestPlayer().getTeamId() == player.getTeamId() &&
		getFastestPlayer().getUniNum() == player.getUniNum())
		return true;
	return false;
}

unsigned InterceptCalculate::getTmmCheckedPlayersNum() const
{
	unsigned tmmCounter = 0;

	for (unsigned i = 0; i < checkedPlayersNum; i++)
		if (getCheckedInterPlayer(i).getPlayer().getTeamId() == TID_TEAMMATE)
			tmmCounter++;

	return tmmCounter;
}

unsigned InterceptCalculate::getOppCheckedPlayersNum() const
{
	unsigned oppCounter = 0;

	for (unsigned i = 0; i < checkedPlayersNum; i++)
		if (getCheckedInterPlayer(i).getPlayer().getTeamId() == TID_OPPONENT)
			oppCounter++;

	return oppCounter;
}

unsigned InterceptCalculate::getCheckedListNum(const Player &player) const
{
	for (unsigned i = 0; i < checkedPlayersNum; i++)
		if (getCheckedInterPlayer(i).getPlayer().getModel() == player.getModel() &&
			getCheckedInterPlayer(i).getPlayer().getTeamId() == player.getTeamId() &&
			getCheckedInterPlayer(i).getPlayer().getUniNum() == player.getUniNum())
				return i;

	return NOVALUE;
}

unsigned InterceptCalculate::getOppPlayersBeforeLine(float x) const
{
	unsigned counter = 0;

	for (unsigned i = 0; i < playersNum; i++)
		if (interceptPlayers[i].getPlayer().getTeamId() == TID_OPPONENT &&
			interceptPlayers[i].getPlayer().getPos().getX() <= x)
				counter++;

	return counter;
}

unsigned InterceptCalculate::getOppPlayersBetweenLines(float x1,
		float x2) const
{
	return abs((int)getOppPlayersBeforeLine(x1) -
			(int)getOppPlayersBeforeLine(x2));
}

const Player &InterceptCalculate::getFastestPlayer() const
{
	return interceptPlayers[fastestPlayerNum].getPlayer();
}

const Player &InterceptCalculate::getFastestTmmPlayer() const
{
	return interceptPlayers[fastestTmmPlayerNum].getPlayer();
}

const Player &InterceptCalculate::getFastestOppPlayer() const
{
	return interceptPlayers[fastestOppPlayerNum].getPlayer();
}

const InterceptPlayer &InterceptCalculate::getFastestInterPlayer() const
{
	return interceptPlayers[fastestPlayerNum];
}

const InterceptPlayer &InterceptCalculate::getFastestTmmInterPlayer() const
{
	return interceptPlayers[fastestTmmPlayerNum];
}

const InterceptPlayer &InterceptCalculate::getFastestOppInterPlayer() const
{
	return interceptPlayers[fastestOppPlayerNum];
}

const InterceptPlayer &InterceptCalculate::getCheckedInterPlayer(unsigned num) const
{
	return interceptPlayers[checkedPlayers[num]];
}

void InterceptCalculate::logCheckedPlayers() const
{
	LOG << "InterceptCalculate checked players:" << endl;

	for (unsigned i = 0; i < checkedPlayersNum; i++)
		LOG << "\t" << i << ": "
			<< getCheckedInterPlayer(i).getPlayer().getTeamId() << " "
			<< getCheckedInterPlayer(i).getPlayer().getUniNum() << " "
			<< getCheckedInterPlayer(i).getInterTime() << endl;
}

void InterceptCalculate::setBlockedBall()
{
	virtualBall.setVel() = blockedBallVel;
	virtualBall.setDecay(1);
}

void InterceptCalculate::getNearestOppDistDirToFastest(float &dist, float &dir)
{
	unsigned miner = NOVALUE;
	float min = 0xFFFF;

	if (existFastestOppPlayer())
	{
		dist = 0;
		dir = 0;

		return;
	}

	for (unsigned i = 0; i < playersNum; i++)
		if (interceptPlayers[i].getPlayer().getTeamId() == TID_OPPONENT)
		{
			float cur = interceptPlayers[i].getPlayer().getDistance(
					getFastestInterPlayer().getInterPoint()) -
					interceptPlayers[i].getDashDist();
			if (miner == NOVALUE || cur < min)
			{
				miner = i;
				min = cur;
			}
		}

	if (miner == NOVALUE)
	{
		dist = NOVALUE;
		dir = NOVALUE;

		return;
	}

	dir = Degree::arcTan2((interceptPlayers[miner].getPlayer().getPos().getY() -
			getFastestInterPlayer().getInterPoint().y),
			(interceptPlayers[miner].getPlayer().getPos().getX() -
			getFastestInterPlayer().getInterPoint().x));

	dist = interceptPlayers[miner].getPlayer().getDistance(
				getFastestInterPlayer().getInterPoint()) -
				interceptPlayers[miner].getDashDist();
}

// Getting functions
bool InterceptCalculate::isLogging() const
{
	return loggingFlag;
}

unsigned InterceptCalculate::getPlayersNum() const
{
	return playersNum;
}

unsigned InterceptCalculate::getFastestPlayerNum() const
{
	return fastestPlayerNum;
}

unsigned InterceptCalculate::getFastestTmmPlayerNum() const
{
	return fastestTmmPlayerNum;
}

unsigned InterceptCalculate::getFastestOppPlayerNum() const
{
	return fastestOppPlayerNum;
}

unsigned InterceptCalculate::getCheckedPlayersNum() const
{
	return checkedPlayersNum;
}

unsigned InterceptCalculate::getFastestCheckedPlayerNum() const
{
	return fastestCheckedPlayerNum;
}

unsigned InterceptCalculate::getFastestTmmCheckedPlayerNum() const
{
	return fastestTmmCheckedPlayerNum;
}

unsigned InterceptCalculate::getFastestOppCheckedPlayerNum() const
{
	return fastestOppCheckedPlayerNum;
}

PlayMode InterceptCalculate::getPlayMode() const
{
	return playMode;
}

const Ball &InterceptCalculate::getVirtualBall() const
{
	return virtualBall;
}

const Player *InterceptCalculate::getKickPlayer() const
{
	return kickPlayer;
}

// Setting functions
void InterceptCalculate::setLogging(bool loggingFlagArg)
{
	loggingFlag = loggingFlagArg;
}

void InterceptCalculate::setKickCycles(unsigned kickCyclesArg)
{
	kickCycles = kickCyclesArg;
}

void InterceptCalculate::setShareCycles(unsigned shareCyclesArg)
{
	shareCycles = shareCyclesArg;
}

void InterceptCalculate::setKickerSeeBallDelay(unsigned kickerSeeBallDelayArg)
{
	kickerSeeBallDelay = kickerSeeBallDelayArg;
}

void InterceptCalculate::setTmmSlowDashPower(float tmmSlowDashPowerArg)
{
	tmmSlowDashPower = tmmSlowDashPowerArg;
}

void InterceptCalculate::setTmmFastDashPower(float tmmFastDashPowerArg)
{
	tmmFastDashPower = tmmFastDashPowerArg;
}

void InterceptCalculate::setTmmDirectDashPower(float tmmDirectDashPowerArg)
{
	tmmDirectDashPower = tmmDirectDashPowerArg;
}

void InterceptCalculate::setTmmGoalieDirectDashPower(float tmmGoalieDirectDashPowerArg)
{
	tmmGoalieDirectDashPower = tmmGoalieDirectDashPowerArg;
}

void InterceptCalculate::setTmmSeeBallDelay(unsigned tmmSeeBallDelayArg)
{
	tmmSeeBallDelay = tmmSeeBallDelayArg;
}

void InterceptCalculate::setTmmGoalieSeeBallDelay(unsigned tmmGoalieSeeBallDelayArg)
{
	tmmGoalieSeeBallDelay = tmmGoalieSeeBallDelayArg;
}

void InterceptCalculate::setTmmPosDevRate(float tmmPosDevRateArg)
{
	tmmPosDevRate = tmmPosDevRateArg;
}

void InterceptCalculate::setOppSlowDashPower(float oppSlowDashPowerArg)
{
	oppSlowDashPower = oppSlowDashPowerArg;
}

void InterceptCalculate::setOppFastDashPower(float oppFastDashPowerArg)
{
	oppFastDashPower = oppFastDashPowerArg;
}

void InterceptCalculate::setOppDirectDashPower(float oppDirectDashPowerArg)
{
	oppDirectDashPower = oppDirectDashPowerArg;
}

void InterceptCalculate::setOppGoalieDirectDashPower(float oppGoalieDirectDashPowerArg)
{
	oppGoalieDirectDashPower = oppGoalieDirectDashPowerArg;
}

void InterceptCalculate::setOppSeeBallDelay(unsigned oppSeeBallDelayArg)
{
	oppSeeBallDelay = oppSeeBallDelayArg;
}

void InterceptCalculate::setOppGoalieSeeBallDelay(unsigned oppGoalieSeeBallDelayArg)
{
	oppGoalieSeeBallDelay = oppGoalieSeeBallDelayArg;
}

void InterceptCalculate::setOppPosDevRate(float oppPosDevRateArg)
{
	oppPosDevRate = oppPosDevRateArg;
}

void InterceptCalculate::setVirtualBall(const Ball &virtualBallArg)
{
	virtualBall = virtualBallArg;
}

void InterceptCalculate::setBlockedBallVel(const Vector &blockedBallVelArg)
{
	blockedBallVel = blockedBallVelArg;
}

InterceptPlayer &InterceptCalculate::setLastInterPlayer()
{
	return interceptPlayers[playersNum - 1];
}

